package samples.door;

import raft.kilavuz.grid.*;
import raft.kilavuz.util.jpct.Avatar;
import raft.kilavuz.runtime.*;
import com.threed.jpct.*;
import java.io.*;
import java.util.*;

/**
 * creates and stores path finding data for door sample. 
 *
 * @author hakan eryargi (r a f t)
 */
public class CreateData {
    
    
    /** loads obejcts from file, rotates them to match jPCT corrdinate system,
     * and finally builds them */
    static Object3D[] loadObjects(String file) {
        Object3D[] objects = Loader.load3DS(file, 1);
        for (Object3D o : objects) {
            // convert 3dmax to jPCT corrdinates
            o.rotateX((float)-Math.PI/2);
            o.rotateMesh();
            o.setRotationMatrix(new Matrix());
            o.build();
        }
        return objects;
    }
    
    
    public static void main(String[] args) throws Exception {
        Logger.setOnError(Logger.ON_ERROR_THROW_EXCEPTION);
        Logger.setLogLevel(Logger.LL_ERRORS_AND_WARNINGS);
        
        World world = new World();
        
        // load and add our map objects and set their collision status
        for (Object3D o : loadObjects("model/level.3ds")) {
            o.setCollisionMode(Object3D.COLLISION_CHECK_OTHERS);
            o.setCollisionOptimization(Object3D.COLLISION_DETECTION_OPTIMIZED);
            world.addObject(o);
        }
        Object3D doorKey = loadObjects("model/door_key.3ds")[0];
        doorKey.setCollisionMode(Object3D.COLLISION_CHECK_OTHERS);
        doorKey.setCollisionOptimization(Object3D.COLLISION_DETECTION_OPTIMIZED);
        world.addObject(doorKey);
        
        // this is a small scene so no need for octrees.
        // for larger scenes they certainly help
        Avatar avatar = new Avatar();
        float[] bb = avatar.getMesh().getBoundingBox();
        
        float cellWidth = Math.max(bb[1]-bb[0], bb[5]-bb[4]) / 2;
        float cellHeight = bb[3]-bb[2];
        
        System.out.println("cellWidth: " + cellWidth + " cellHeight: " + cellHeight);
        
        float stepRatio = 0.3f;
        float jumpRatio = 0; // we dont want our avatar to jump
        float fallRatio = 1f; // can safely fall down as much as his height
        
        SimpleStepChecker stepChecker = new SimpleStepChecker(world, cellWidth, cellHeight,
                stepRatio, jumpRatio, fallRatio);
        
        Vector3 doorExitPos = new Vector3(loadObjects("model/door_exit_cell.3ds")[0].getCenter());
        Vector3 doorKeyPos = new Vector3(loadObjects("model/door_key_cell.3ds")[0].getCenter());

        Vector3 startPos = new Vector3(loadObjects("model/start_cell.3ds")[0].getCenter());
        float distance = stepChecker.getDistanceToGround(startPos);
        if (distance == SimpleStepChecker.NO_GROUND)
            throw new IllegalStateException("no ground under: " + startPos);
        startPos.y += distance;
        
        // we are ready, here we go
        FloodFill floodFill = new FloodFill(startPos, stepChecker, cellWidth, cellHeight, 
                0.2f, stepRatio, jumpRatio, fallRatio, 0, FloodFill.DIAGONAL_JUMP_ALLOWED);
        
        Grid grid = floodFill.getGrid();

        // we intervene with sector creation and explicitly create a sector 
        // which has just the same shape and bounds with the door
        float[] doorBounds = loadObjects("model/door.3ds")[0].getMesh().getBoundingBox();
        Grid.Sector doorSector = grid.createSector(grid.getCellsAt(doorBounds));
        
        // let grid create the rest of sectors
        grid.sectorize();
        
        // our positions are on the ground, it's always a good idea 
        // to move ground positions a little up to avoid rounding errors
        doorExitPos.y -= cellHeight / 5;
        doorKeyPos.y -= cellHeight / 5;
        
        Grid.Cell doorExitCell = grid.getCellAt(doorExitPos);
        Grid.Cell doorKeyCell = grid.getCellAt(doorKeyPos);
        // we create a portal between door key and and exit point of door
        Grid.Portal portal = grid.createPortal(doorKeyCell, doorExitCell, DoorSample.ACTION_DOOR);
        
        // create the land
        Land land = new Land(grid);
        
        // portals and sectors always have same id's in Grid and Land
        List<Land.Portal> portalsToDoor = land.getSector(doorSector.id).getArrivingPortals();
        for (Land.Portal p : portalsToDoor)
            p.setEnabled(false);
        Land.Portal doorPortal = land.getPortal(portal.id);
        // we will use this piece is information in door sample
        // so it's a good idea to store it with our land
        land.setUserObject(new DoorSample.DoorPortals(doorPortal, portalsToDoor));
                
        // that's all now save data for future reference
        grid.saveToStream(new FileOutputStream("data/grid.ser"));
        System.out.println("saved grid to data/grid.ser");
        
        land.saveToStream(new FileOutputStream("data/land.ser"));
        System.out.println("saved land to data/land.ser");
    }
    
}
