package samples.door;

import raft.kilavuz.runtime.*;
import com.threed.jpct.*;

import java.util.*;
import java.util.List;
import java.io.*;
import javax.swing.*;
import java.awt.*;
import java.awt.event.*;
import javax.sound.sampled.*;
import raft.kilavuz.util.jpct.Avatar;
import raft.kilavuz.util.jpct.RenderPanel;

/**
 * this sample demonstrates how a door can be implemented on top of kilavuz.
 *
 * @author  hakan eryargi (r a f t)
 */
public class DoorSample implements RenderPanel.Listener {
    
    /** id of door action */
    public static final short ACTION_DOOR = 200;
    
    private final AStar astar = new AStar(); // pathfinder
    private PathIterator path = null; // the path avatar will follow
    private final float gravity; // gravity constant in world units
    
    private final RenderPanel renderPanel; // the swing panel actually does the rendering
    private final World world; // world instance
    private boolean freeCamera = true; // is camera free or following avatar
    
    private final Avatar avatar; // the avatar
    private final Door door; // the door
    private final Land land; // runtime pathfinding information

    // portalPasser creates necessary path segments for passing through door and falldown
    private final PortalPasser portalPasser = new PortalPasserImpl();
    
    // data structure holding informatio door related portals
    private final DoorPortals doorPortals; 
    private final SimpleVector startPos; // start point of avatar
    
    // context encapsulates information about pathfinding information
    private final PathContext context;
    
    // we use path markers to visualize the path to the user
    private List<Object3D> pathMarkers = new ArrayList<Object3D>();
    // original marker object, all others copies are created by clonig this
    private final Object3D marker;
    
    private Clip doorSound; // played when door is opened
    
    /** creates the sample */
    public DoorSample() throws Exception {
        this.world = new World();
        
        // laod previously saved runtime pathfinding information
        this.land = Land.loadFromStream(getResource("/data/land.ser"));
        this.doorPortals = (DoorPortals) land.getUserObject();
        
        // load map objects and set their collision status
        // setting collision status is necessary for moving camera
        for (Object3D o : loadObjects("/model/level.3ds")) {
            o.setCollisionMode(Object3D.COLLISION_CHECK_OTHERS);
            o.setCollisionOptimization(Object3D.COLLISION_DETECTION_OPTIMIZED);
            world.addObject(o);
        }
        // door key is the button on the ground which opens the door
        Object3D doorKey = loadObjects("/model/door_key.3ds")[0];
        doorKey.setCollisionMode(Object3D.COLLISION_CHECK_OTHERS);
        doorKey.setCollisionOptimization(Object3D.COLLISION_DETECTION_OPTIMIZED);
        world.addObject(doorKey);
        
        // create the avatar
        this.avatar = new Avatar();
        float[] bb = avatar.getMesh().getBoundingBox();
        final float avatarHeight = bb[3]-bb[2];
        world.addObject(avatar);
        
        // create the path context
        this.context = new PathContext();
        context.distanceToEdges = 1.5f; // our avatar will try to stay this much away from sides
        context.velocity = avatarHeight / land.cellWidth; // as much as its height per second
        context.turnVelocity = (float)(Math.PI / 2f); // radians per second
        this.gravity = avatarHeight * 10;
        
        // load the marker object
        Texture flagTexture = new Texture(getResource("/textures/flag.jpg"));
        TextureManager.getInstance().addTexture("FLAG.JPG", flagTexture);
        this.marker = loadObjects("/model/flag.3ds")[0];
        marker.setBillboarding(true);
        
        // load the door object
        Object3D doorOb = loadObjects("/model/door.3ds")[0];
        world.addObject(doorOb);
        
        this.door = new Door(doorOb, loadObjects("/model/door_key_cell.3ds")[0], avatarHeight*3/2, avatarHeight*2);
        
        // these seems to be good positions to see the all scene
        // we run the sample once, moved camera around and got these values
        SimpleVector cameraPos = new SimpleVector(770, -1115, -140);
        SimpleVector cameraDir = new SimpleVector(-0.75, 0.65, 0);
        
        world.getCamera().setPosition(cameraPos);
        world.getCamera().setBack(cameraDir.getRotationMatrix().invert3x3());
        
        startPos = land.getPoint(loadObjects("/model/start_cell.3ds")[0].getCenter()).getLocation();
        avatar.translate(startPos);
        
        // add a few lights to scene
        world.setAmbientLight(64, 64, 64);
        for (Object3D o : loadObjects("/model/lights.3ds")) {
            world.addLight(o.getCenter(), Color.LIGHT_GRAY);
        }

        // laod the audio clip
        try {
            AudioInputStream audioStream = AudioSystem.getAudioInputStream(new BufferedInputStream(new FileInputStream("sounds/door.wav")));
            doorSound = (Clip) AudioSystem.getLine(new DataLine.Info(Clip.class, audioStream.getFormat()));
            doorSound.open(audioStream);
        } catch (Exception e) {
            // ok, we can live without sound
            e.printStackTrace();
        }
        
        // cameraEllipsoid will be used by renderPanel to move camera around
        SimpleVector cameraEllipsoid = new SimpleVector(avatarHeight, avatarHeight, avatarHeight);
        cameraEllipsoid.scalarMul(0.2f);
        // create the RenderPanel which actually does the rendering
        renderPanel = new RenderPanel(world, land, this, cameraEllipsoid, 2 * avatarHeight);
        renderPanel.setPreferredSize(new Dimension(640, 480));
        
        
        renderPanel.addFocusListener(new FocusListener() {
            public void focusGained(FocusEvent e) {
                paused = false;
            }
            public void focusLost(FocusEvent e) {
                paused = true;
            }
        }); 
        
        renderPanel.addKeyListener(new KeyAdapter() {
            public void keyPressed(KeyEvent e) {
                switch (e.getKeyCode()) {
                    case KeyEvent.VK_ESCAPE:
                        running = false;
                        System.exit(0);
                        break;
                    case KeyEvent.VK_R:
                        reset();
                        break;
                    case KeyEvent.VK_F:
                        freeCamera = !freeCamera;
                        renderPanel.followAvatar(freeCamera ? null : avatar, 
                                avatarHeight*2/3, avatarHeight*2, (float)Math.toRadians(10));
                        break;
                }
            }
        });
        
    }
    
    /** clears path and moves avatar to start position */
    private void reset() {
        clearPathMarkers();
        path = null;
        
        avatar.getTranslationMatrix().setIdentity();
        avatar.translate(land.getPoint(startPos).getLocation());
    }
    
    /** moves the avatar along the path if any. called every frame. */
    private void moveAvatar(float seconds) {
        if (path != null) {
            if (path.hasNext()) {
                PathPosition position = path.next(seconds);
                avatar.getTranslationMatrix().setIdentity();
                avatar.translate(position.location);
                //avatar.getRotationMatrix().setTo(position.direction.getRotationMatrix());
                avatar.setRotationMatrix(position.direction.getRotationMatrix());
                //position.direction.getRotationMatrix().scalarMul()
                avatar.step(position.animation, seconds);
            } else {
                renderPanel.showMessage("path ended", 1000);
                path = null;
                clearPathMarkers();
            }
        }
        // this method completes animation cycles of avatar, if any
        avatar.frameCompleted(seconds);
    }
    
    /** render panel notifies us when user clicked on a collidable object */
    public void mouseClickedAt(SimpleVector vector) {
        Land.Point from = land.getPoint(avatar.getTransformedCenter());
        Land.Point to = land.getPoint(vector);
        
        findPath(from, to);
    }
    
    /** finds the path between given points */
    private void findPath(Land.Point from, Land.Point to)  {
        try {
            if (from == null || to == null)
                return;
            
            long start = System.currentTimeMillis();
            AStar.Path astarPath = astar.findPath(from.sector, to.sector, context);
            renderPanel.showMessage("path found in " +  (System.currentTimeMillis() - start) + " millis", 2000);
            
            clearPathMarkers();
            
            // create the path points which forms corners of path
            List<Land.Point> pathPoints = Paths.pathPoints(from, astarPath, to,
                    Paths.PORTAL_EXITS_DONT_CREATE);
            
            // place markers on path points
            // skip first one, as its where we stand
            for (int i = 1; i < pathPoints.size(); i++) {
                Land.Point pathPoint = pathPoints.get(i);
                Object3D newMarker = marker.cloneObject();
                newMarker.getTranslationMatrix().setIdentity();
                newMarker.translate(pathPoint.getLocation());
                world.addObject(newMarker);
                pathMarkers.add(newMarker);
            }
            // create a curved path
            // note this method uses PathContext encapsulated in AStar.Path
            this.path = Paths.curvedPath(from, astarPath, to,
                    avatar.getRotationMatrix().getZAxis(), null, portalPasser);
            
        } catch (NoPathException nse) {
            renderPanel.showMessage("no path found", 2000);
            clearPathMarkers();
        }
    }
    
    /** removes previously placed ((if any) path markers */
    private void clearPathMarkers() {
        for (Object3D marker : pathMarkers)
            world.removeObject(marker.getID());
        pathMarkers.clear();
    }
    
    long frameNo = 0;
    long lastTime;
    long excessTime = 0;
    final long granularity = 25; // milli seconds
    
    boolean running = false;
    boolean paused = false;
    
    private void gameLoop() {
        running = true;
        
        lastTime = System.currentTimeMillis();
        excessTime = 0;
        
        final float seconds = granularity / 1000f;
        
        while(running) {
            try {
                // as RenderPanel.render should only be called from awt thread
                // we submit all code to awt thread
                SwingUtilities.invokeAndWait(new Runnable() { public void run() {
                    long now = System.currentTimeMillis();
                    excessTime += (now - lastTime);
                    lastTime = now;
                    
                    while(excessTime > granularity) {
                        frameNo++;
                        excessTime -= granularity;
                        
                        door.move(seconds);
                        moveAvatar(seconds);
                        renderPanel.placeCamera(seconds);
                    }
                    if (paused) {
                        try {
                            Thread.sleep(granularity);
                        } catch (InterruptedException e) {}
                    } else {
                        renderPanel.render();
                        Thread.yield();
                    }
                }});
            } catch (Exception e) {
                e.printStackTrace();
            }
        } // while running
    }
    
    
    private InputStream getResource(String resource) {
        InputStream in = getClass().getResourceAsStream(resource);
        if (in == null)
            throw new MissingResourceException(resource, null, null);
        return in;
    }
    
    /** loads obejcts from file, rotates them to match jPCT corrdinate system,
     * and finally builds them */
    private Object3D[] loadObjects(String resource) {
        Object3D[] objects = Loader.load3DS(getResource(resource), 1);
        for (Object3D o : objects) {
            // convert 3dmax to jPCT corrdinates
            o.rotateX((float)-Math.PI/2);
            o.rotateMesh();
            o.setRotationMatrix(new Matrix());
            o.build();
        }
        return objects;
    }
    
    
    /** a small class which handles door movement */
    class Door {
        final Object3D object3d; // the 3d object for door 
        // the center of an imaginary point over key, we use it to detect if avatar is over key
        final SimpleVector keyCenter; 
        final SimpleVector keyEllipsoid; // bounds of that imaginary point
        final float maxHeight; // how high door will go
        final float speed; // speed of door
        float currentHeight = 0;
        long keyDisabledTill = 0;
        boolean open = false; // is door open
        boolean opening = false; // is it opening at the moment
        boolean closing = false; // is it closing at the moment
        
        /** create the door */
        Door(Object3D object3d, Object3D keyMarker, float maxHeight, float speed) {
            this.object3d = object3d;
            this.keyCenter = keyMarker.getCenter();
            float[] bb = keyMarker.getMesh().getBoundingBox();
            this.keyEllipsoid = new SimpleVector(bb[1]-bb[0], bb[3]-bb[2], bb[5]-bb[4]);
            this.maxHeight = maxHeight;
            this.speed = speed;
        }
        /** opens the door if it's not open or opening */
        private void open() {
            if (open || opening)
                return;
            
            opening = true;
            if (doorSound != null) {
                doorSound.setFramePosition(0);
                doorSound.start();
            }
        }
        
        /** actually moves the door. called every frame */
        private void move(float seconds) {
            if (opening) {
                currentHeight += seconds * speed;
                if (currentHeight > maxHeight) {
                    currentHeight = maxHeight;
                    opening = false;
                    open = true;
                    doorPortals.doorIsOpen();
                }
                object3d.getTranslationMatrix().setIdentity();
                object3d.translate(0, -currentHeight, 0);
            } else if (closing) {
                currentHeight -= seconds * speed;
                if (currentHeight < 0) {
                    currentHeight = 0;
                    closing = false;
                    open = false;
                    doorPortals.doorIsClosed();
                }
                object3d.getTranslationMatrix().setIdentity();
                object3d.translate(0, -currentHeight, 0);
            } else {
                if (System.currentTimeMillis() < keyDisabledTill)
                    return;
                
                if (avatar.ellipsoidIntersectsAABB(keyCenter, keyEllipsoid)) {
                    keyDisabledTill = System.currentTimeMillis() + 5000;
                    
                    if (doorSound != null) {
                        doorSound.setFramePosition(0);
                        doorSound.start();
                    }
                    
                    if (open)
                        closing = true;
                    else opening = true;
                    
                }
            }
        }
    }
    
    /** creates path segments to pass through door portal and fall down segments */
    class PortalPasserImpl implements PortalPasser {
        public PathIterator getPathIterator(Land.Portal portal, Vector3 from, Vector3 to, Vector3 fromDirection) {
            switch (portal.getAction()) {
                case ACTION_DOOR:
                    return createDoorIterator(from, to, fromDirection);
                case Land.ACTION_FALLDOWN:
                    return Paths.acceleratedSegment(fromDirection, from, to, Paths.ANIMATE_FALLDOWN, 0, gravity);
                default:
                    return null;
            }
        };
        /** creates the PathIterator to pass through door portal */
        private PathIterator createDoorIterator(Vector3 from, Vector3 to, Vector3 fromDirection) {
            try {
                // remember door portal was from the door key to a point behind the door
                // we will create a PathIterator between the two points
                
                // disable door portal and enable the portals at the actual door
                doorPortals.doorKeyPortal.setEnabled(false);
                for (Land.Portal portal : doorPortals.toDoorPortals)
                    portal.setEnabled(true);
                
                Land.Point fromPoint = land.getPoint(from);
                Land.Point toPoint = land.getPoint(to);
                // pathfinder will find us a short path between two points
                AStar.Path doorPath = astar.findPath(fromPoint.sector, toPoint.sector, context);
                
                // restore the portals' states
                doorPortals.doorKeyPortal.setEnabled(true);
                for (Land.Portal portal : doorPortals.toDoorPortals)
                    portal.setEnabled(false);
                
                // as there should be nothing special at the path segment 
                // we can use Paths to actually create an iterator
                PathIterator path = Paths.curvedPath(fromPoint, doorPath, toPoint,
                        fromDirection, null, portalPasser);
                
                // we are wrapping the path just to open the door when avatar reaches key
                PathIterator wrapper = new Paths.Wrapper(path) {
                    
                    boolean opened = false;
                    public PathPosition next(float time) {
                        // opening once is enogh
                        if (!opened) {
                            door.open();
                            opened = true;
                        }
                        return super.next(time);
                    }
                };
                // that's it, returned the wrapped iterator
                return wrapper;
                
            } catch (NoPathException npe) {
                // this shouldnt happen
                return null;
            }
        }
    }
    
    /** a small class to hold information about door related portals */
    public static class DoorPortals implements java.io.Serializable {
        private static final long serialVersionUID = 1;
        
        // this is the portal we created from door key to door exit
        public final Land.Portal doorKeyPortal;
        // these are the portals which lead to door area
        public final List<Land.Portal> toDoorPortals;
        
        public DoorPortals(Land.Portal doorKeyPortal, List<Land.Portal> toDoorPortals) {
            this.doorKeyPortal = doorKeyPortal;
            this.toDoorPortals = Collections.unmodifiableList(toDoorPortals);
        }
        
        /** adjust the portals state so that door is open */
        private void doorIsOpen() {
            doorKeyPortal.setEnabled(false);
            for (Land.Portal portal : toDoorPortals)
                portal.setEnabled(true);
        }
        
        /** adjust the portals state so that door is closed */
        private void doorIsClosed() {
            doorKeyPortal.setEnabled(true);
            for (Land.Portal portal : toDoorPortals)
                portal.setEnabled(false);
        }
    }
    
    public static void main(String[] args) throws Exception {
        Logger.setOnError(Logger.ON_ERROR_THROW_EXCEPTION);
        Logger.setLogLevel(Logger.LL_ERRORS_AND_WARNINGS);
        
        // set a few jPCT config variables
        Config.farPlane = 10000;
        Config.linearDiv = 200;
        Config.lightMul = 5;
        
        // create the sample
        DoorSample sample = new DoorSample();
        
        // pack it in JFrame
        JFrame frame = new JFrame("kilavuz - door sample");
        frame.add(sample.renderPanel);
        frame.pack();
        frame.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
        frame.setVisible(true);
        
        sample.renderPanel.requestFocus();
        
        // here we go
        sample.gameLoop();
    }
    
}
