package samples.elevator;

import raft.kilavuz.grid.*;
import raft.kilavuz.runtime.*;
import raft.kilavuz.util.jpct.*;

import com.threed.jpct.*;
import java.io.*;
import java.util.*;

/**
 * creates and stores path finding data for elevator sample.
 *
 * @author hakan eryargi (r a f t)
 */
public class CreateData {
    
    
    /** loads obejcts from file, rotates them to match jPCT corrdinate system,
     * and finally builds them */
    static Object3D[] loadObjects(String file) {
        Object3D[] objects = Loader.load3DS(file, 1);
        for (Object3D o : objects) {
            // convert 3dmax to jPCT corrdinates
            o.rotateX((float)-Math.PI/2);
            o.rotateMesh();
            o.setRotationMatrix(new Matrix());
            o.build();
        }
        return objects;
    }
    
    
    public static void main(String[] args) throws Exception {
        Logger.setOnError(Logger.ON_ERROR_THROW_EXCEPTION);
        Logger.setLogLevel(Logger.LL_ERRORS_AND_WARNINGS);
        
        World world = new World();
        
        // load and add our map objects and set their collision status
        for (Object3D o : loadObjects("model/level.3ds")) {
            o.setCollisionMode(Object3D.COLLISION_CHECK_OTHERS);
            o.setCollisionOptimization(Object3D.COLLISION_DETECTION_OPTIMIZED);
            world.addObject(o);
        }
        
        // this is a small scene so no need for octrees.
        // for larger scenes they certainly help
        
        
        Avatar avatar = new Avatar();
        float[] bb = avatar.getMesh().getBoundingBox();
        
        float cellWidth = Math.max(bb[1]-bb[0], bb[5]-bb[4]) / 2;
        float cellHeight = bb[3]-bb[2];
        
        System.out.println("cellWidth: " + cellWidth + " cellHeight: " + cellHeight);
        
        float stepRatio = 0.3f;
        float jumpRatio = 0; // we dont want our avatar to jump
        float fallRatio = 1f; // can safely fall down as much as his height
        
        SimpleStepChecker stepChecker = new SimpleStepChecker(world, cellWidth, cellHeight,
                stepRatio, jumpRatio, fallRatio);
        
        Vector3 downPos = new Vector3(loadObjects("model/downcell.3ds")[0].getCenter());
        Vector3 upPos = new Vector3(loadObjects("model/upcell.3ds")[0].getCenter());
        
        float distance = stepChecker.getDistanceToGround(downPos);
        if (distance == SimpleStepChecker.NO_GROUND)
            throw new IllegalStateException("no ground under: " + downPos);
        downPos.y += distance;
        
        distance = stepChecker.getDistanceToGround(upPos);
        if (distance == SimpleStepChecker.NO_GROUND)
            throw new IllegalStateException("no ground under: " + upPos);
        upPos.y += distance;
        
        // normally starting from upPos is sufficient but as that point
        // is close to map side we bump into license limits. 
        // so we start with a point around center of map
        // note we do not need to specify a point downstairs since floodfill will 
        // fall down to downstairs
        Vector3 floodFillStartPos = new Vector3(upPos.x - 5*cellWidth, upPos.y, upPos.z + 3*cellWidth);
        
        // we are ready, here we go
        FloodFill floodFill = new FloodFill(floodFillStartPos, stepChecker, cellWidth, cellHeight, 0.5f,
                stepRatio, jumpRatio, fallRatio, 0, FloodFill.DIAGONAL_JUMP_ALLOWED);
        
        // flood fill completed, now get the grid 
        Grid grid = floodFill.getGrid();
        // we don not create any special sector in this example, 
        // so let grid do it all automatically
        grid.sectorize();
        
        // our positions are on the ground, it's always a good idea
        // to move ground positions a little up to avoid rounding errors
        downPos.y -= cellHeight / 5;
        upPos.y -= cellHeight / 5;
        
        Grid.Cell downCell = grid.getCellAt(downPos);
        Grid.Cell upCell = grid.getCellAt(upPos);
        
        // define a portal between down elevation position and up elevator position
        Grid.Portal portal = grid.createPortal(downCell, upCell, ElevatorSample.ACTION_ELEVATOR);
        // set a high cost for elevator portal, 
        // as we want our avatar to use elevator only when necessary
        portal.setExplicitCost(new Float(1000));
        
        // that's all now save data for future reference
        grid.saveToStream(new FileOutputStream("data/grid.ser"));
        System.out.println("saved grid to data/grid.ser");
        
        Land land = new Land(grid);
        land.saveToStream(new FileOutputStream("data/land.ser"));
        System.out.println("saved land to data/land.ser");
    }
    
}
