package samples.elevator;

import raft.kilavuz.runtime.*;
import raft.kilavuz.util.jpct.*;
import com.threed.jpct.*;

import java.util.*;
import java.util.List;
import java.io.*;
import javax.swing.*;
import java.awt.*;
import java.awt.event.*;

/**
 * this sample demonstrates how an elevator can be implemented on top of kilavuz.
 * it also shows the usage of Paths.Condition and Paths.Wrapper
 *
 * @author  hakan eryargi (r a f t)
 */
public class ElevatorSample implements RenderPanel.Listener {
    /** id of elevator action */
    public static final short ACTION_ELEVATOR = 100;
    
    private final AStar astar = new AStar(); // pathfinder
    private PathIterator path = null; // the path avatar will follow
    private final float gravity; // gravity constant in world units
    
    private final RenderPanel renderPanel; // the swing panel actually does the rendering
    private final World world; // world instance
    private boolean freeCamera = true; // is camera free or following avatar
    
    private final Avatar avatar; // the avatar
    private final Elevator elevator; // the elevator
    private final Land land; // runtime pathfinding information
    private final SimpleVector elevatorSurface; // the point on center of elevator when it's downstairs
    
    // portalPasser creates necessary path segments for elevator and falldown
    private final PortalPasser portalPasser = new PortalPasserImpl();
    private final SimpleVector startPos; // start point of avatar
    
    // we use path markers to visualize the path to the user
    private List<Object3D> pathMarkers = new ArrayList<Object3D>();
    // original marker object, all others copies are created by clonig this
    private final Object3D marker;
    
    // context encapsulates information about pathfinding information
    private final PathContext context;
    
    
    /** creates the sample */
    public ElevatorSample() throws Exception {
        this.world = new World();
        
        // laod previously saved runtime pathfinding information
        this.land = Land.loadFromStream(getResource("/data/land.ser"));
        
        // load map objects and set their collision status
        // setting collision status is necessary for moving camera
        for (Object3D o : loadObjects("/model/level.3ds")) {
            o.setCollisionMode(Object3D.COLLISION_CHECK_OTHERS);
            o.setCollisionOptimization(Object3D.COLLISION_DETECTION_OPTIMIZED);
            world.addObject(o);
        }
        // create the avatar
        this.avatar = new Avatar();
        float[] bb = avatar.getMesh().getBoundingBox();
        final float avatarHeight = bb[3]-bb[2];
        world.addObject(avatar);
        
        // create the path context
        this.context = new PathContext();
        context.distanceToEdges = 1.5f; // our avatar will try to stay this much away from sides
        context.velocity = avatarHeight / land.cellWidth; // as much as its height per second
        context.turnVelocity = (float)(Math.PI / 2f); // radians per second
        this.gravity = avatarHeight * 10;
        
        // load the marker object
        Texture flagTexture = new Texture(getResource("/textures/flag.jpg"));
        TextureManager.getInstance().addTexture("FLAG.JPG", flagTexture);
        this.marker = loadObjects("/model/flag.3ds")[0];
        marker.setBillboarding(true);
        
        // load the elevator object
        Object3D elevatorOb = loadObjects("/model/elevator.3ds")[0];
        world.addObject(elevatorOb);
        
        // these two are the positions where we wait get on and off elevator
        SimpleVector downPos = loadObjects("/model/downcell.3ds")[0].getCenter();
        SimpleVector upPos = loadObjects("/model/upcell.3ds")[0].getCenter();
        
        elevatorSurface = elevatorOb.getCenter();
        // adjust the elevatorSurface y coordinate (elevation) to match with ground elevation
        elevatorSurface.y = land.getPoint(downPos).getElevation();
        
        // elevator will go this much up
        float elevationHeight = elevatorSurface.y - land.getPoint(upPos).getElevation();
        this.elevator = new Elevator(elevatorOb, elevationHeight, 2, avatarHeight / 2f);
        
        // these seems to be good positions to see the all scene
        // we run the sample once, moved camera around and got these values
        SimpleVector cameraPos = new SimpleVector(-740, -800, -1040);
        SimpleVector cameraDir = new SimpleVector(0.4, 0.5, 0.7);
        
        world.getCamera().setPosition(cameraPos);
        world.getCamera().setBack(cameraDir.getRotationMatrix().invert3x3());
        
        // move avatar to start position
        startPos = land.getPoint(downPos).getLocation();
        avatar.translate(startPos);
        
        // add a few lights to scene
        world.setAmbientLight(128, 128, 128);
        for (Object3D o : loadObjects("/model/lights.3ds")) {
            world.addLight(o.getCenter(), Color.BLUE);
        }
        
        // cameraEllipsoid will be used by renderPanel to move camera around
        SimpleVector cameraEllipsoid = new SimpleVector(avatarHeight, avatarHeight, avatarHeight);
        cameraEllipsoid.scalarMul(0.2f);
        // create the RenderPanel which actually does the rendering
        renderPanel = new RenderPanel(world, land, this, cameraEllipsoid, 2 * avatarHeight);
        renderPanel.setPreferredSize(new Dimension(640, 480));
        
        renderPanel.addFocusListener(new FocusListener() {
            public void focusGained(FocusEvent e) {
                paused = false;
            }
            public void focusLost(FocusEvent e) {
                paused = true;
            }
        });
        
        renderPanel.addKeyListener(new KeyAdapter() {
            public void keyPressed(KeyEvent e) {
                switch (e.getKeyCode()) {
                    case KeyEvent.VK_ESCAPE:
                        running = false;
                        System.exit(0);
                        break;
                    case KeyEvent.VK_R:
                        reset();
                        break;
                    case KeyEvent.VK_F:
                        freeCamera = !freeCamera;
                        renderPanel.followAvatar(freeCamera ? null : avatar,
                                avatarHeight*2/3, avatarHeight*2, (float)Math.toRadians(10));
                        break;
                }
            }
        });
        
    }
    /** clears path and moves avatar to start position */
    private void reset() {
        clearPathMarkers();
        path = null;
        
        avatar.getTranslationMatrix().setIdentity();
        avatar.translate(land.getPoint(startPos).getLocation());
    }
    
    /** moves the avatar along the path if any. called every frame. */
    private void moveAvatar(float seconds) {
        if (path != null) {
            if (path.hasNext()) {
                PathPosition position = path.next(seconds);
                avatar.getTranslationMatrix().setIdentity();
                avatar.translate(position.location);
                avatar.setRotationMatrix(position.direction.getRotationMatrix());
                avatar.step(position.animation, seconds);
            } else {
                renderPanel.showMessage("path ended", 1000);
                path = null;
                clearPathMarkers();
            }
        }
        // this method completes animation cycles of avatar, if any
        avatar.frameCompleted(seconds);
    }
    /** render panel notifies us when user clicked on a collidable object */
    public void mouseClickedAt(SimpleVector vector) {
        // get the point in grid space for avatar's current location
        Land.Point from = land.getPoint(avatar.getTransformedCenter());
        // get the point in grid space for clicked point
        Land.Point to = land.getPoint(vector);
        
        findPath(from, to);
    }
    /** finds the path between given points */
    private void findPath(Land.Point from, Land.Point to)  {
        try {
            if (from == null || to == null)
                return;
            
            long start = System.currentTimeMillis();
            AStar.Path astarPath = astar.findPath(from.sector, to.sector, context);
            renderPanel.showMessage("path found in " +  (System.currentTimeMillis() - start) + " millis", 2000);
            
            clearPathMarkers();
            
            // create the path points which forms corners of path
            List<Land.Point> pathPoints = Paths.pathPoints(from, astarPath, to,
                    Paths.PORTAL_EXITS_DONT_CREATE);
            
            // place markers on path points
            // skip first one, as its where we stand
            for (int i = 1; i < pathPoints.size(); i++) {
                Land.Point pathPoint = pathPoints.get(i);
                Object3D newMarker = marker.cloneObject();
                newMarker.getTranslationMatrix().setIdentity();
                newMarker.translate(pathPoint.getLocation());
                world.addObject(newMarker);
                pathMarkers.add(newMarker);
            }
            // create a curved path
            // note this method uses PathContext encapsulated in AStar.Path
            this.path = Paths.curvedPath(from, astarPath, to,
                    avatar.getRotationMatrix().getZAxis(), null, portalPasser);
            
        } catch (NoPathException nse) {
            renderPanel.showMessage("no path found", 2000);
            clearPathMarkers();
        }
    }
    /** removes previously placed ((if any) path markers */
    private void clearPathMarkers() {
        for (Object3D marker : pathMarkers)
            world.removeObject(marker.getID());
        pathMarkers.clear();
    }
    
    private long lastTime;
    private long excessTime = 0;
    private final long granularity = 25; // milli seconds
    
    private boolean running = false;
    private boolean paused = false;
    
    private void gameLoop() {
        running = true;
        
        lastTime = System.currentTimeMillis();
        excessTime = 0;
        
        final float seconds = granularity / 1000f;
        
        while(running) {
            try {
                // as RenderPanel.render should only be called from awt thread
                // we submit all code to awt thread
                SwingUtilities.invokeAndWait(new Runnable() { public void run() {
                    long now = System.currentTimeMillis();
                    excessTime += (now - lastTime);
                    lastTime = now;
                    
                    while(excessTime > granularity) {
                        excessTime -= granularity;
                        
                        elevator.move(seconds);
                        moveAvatar(seconds);
                        renderPanel.placeCamera(seconds);
                    }
                    if (paused) {
                        try {
                            Thread.sleep(granularity);
                        } catch (InterruptedException e) {}
                    } else {
                        renderPanel.render();
                        Thread.yield();
                    }
                }});
            } catch (Exception e) {
                e.printStackTrace();
            }
        } // while running
    }
    
    
    private InputStream getResource(String resource) {
        InputStream in = getClass().getResourceAsStream(resource);
        if (in == null)
            throw new MissingResourceException(resource, null, null);
        return in;
    }
    
    /** loads obejcts from file, rotates them to match jPCT corrdinate system,
     * and finally builds them */
    private Object3D[] loadObjects(String resource) {
        Object3D[] objects = Loader.load3DS(getResource(resource), 1);
        for (Object3D o : objects) {
            // convert 3dmax to jPCT corrdinates
            o.rotateX((float)-Math.PI/2);
            o.rotateMesh();
            o.setRotationMatrix(new Matrix());
            o.build();
        }
        return objects;
    }
    
    /** a small class which handles elevator movement */
    class Elevator {
        final Object3D object3d; // the 3d object for elevator
        final float maxHeight; // how high elevator will go
        final float waitSeconds; // at up and down stairs how long will it wait
        final float speed; // speed of elevator
        float currentHeight = 0; 
        float timeLeft = 0; // time left to move
        boolean goingUp; // going up or down
        
        /** create the elevator */
        Elevator(Object3D object3d, float maxHeight, float waitSeconds, float speed) {
            this.object3d = object3d;
            this.maxHeight = maxHeight;
            this.waitSeconds = waitSeconds;
            this.speed = speed;
            timeLeft = waitSeconds;
            goingUp = true;
        }
        /** actually moves the elevator. called every frame */
        private void move(float seconds) {
            if (timeLeft > 0) {
                timeLeft -= seconds;
            } else {
                if (goingUp) {
                    currentHeight += seconds * speed;
                    if (currentHeight > maxHeight) {
                        currentHeight = maxHeight;
                        timeLeft = waitSeconds;
                        goingUp = false;
                    }
                } else {
                    currentHeight -= seconds * speed;
                    if (currentHeight < 0) {
                        currentHeight = 0;
                        timeLeft = waitSeconds;
                        goingUp = true;
                    }
                }
                object3d.getTranslationMatrix().setIdentity();
                object3d.translate(0, -currentHeight, 0);
            }
        }
        
        boolean isAtDownstairs() {
            return (currentHeight == 0);
        }
        boolean isAtUpstairs() {
            return (currentHeight == maxHeight);
        }
    }
    /** creates path segments to pass through elevator and fall down segments */
    class PortalPasserImpl implements PortalPasser {
        public PathIterator getPathIterator(Land.Portal portal, Vector3 from, Vector3 to, Vector3 fromDirection) {
            switch (portal.getAction()) {
                case ACTION_ELEVATOR:
                    // elevator movement is a bit complicated so we will 
                    // create it in another method
                    return createElevatorIterator(from, to, fromDirection);
                case Land.ACTION_FALLDOWN:
                    // falling down is an accelerated movement right ? with acceleration = gravity
                    return Paths.acceleratedSegment(fromDirection, from, to, Paths.ANIMATE_FALLDOWN, 0, gravity);
                default:
                    return null;
            }
        };
        
        private PathIterator createElevatorIterator(Vector3 from, Vector3 to, Vector3 fromDirection) {
            // we will turn to elevator, wait for elevator to come, 
            // get on, go up, turn while going up, and get off. 
            // all of these are segments of our path
            List<PathIterator> segments = new ArrayList<PathIterator>();
            
            // turn to elevator
            SimpleVector firstDirection = elevatorSurface.calcSub(from);
            PathIterator turnToElevator = Paths.turnSegment(fromDirection, firstDirection, from, context.turnVelocity);
            segments.add(turnToElevator);
            
            // we will wait till elevator comes down
            // we will do it by using a Paths.Condition
            Paths.Condition waitForComesDown = new Paths.Condition() {
                public boolean isConditionMet() {
                    // wait for elevator to come
                    if (!elevator.isAtDownstairs())
                        return false;
                    // if there is not much time to board, wait for second come
                    if (elevator.timeLeft < 0.5)
                        return false;
                    return true;
                }
            };
            // we use the last position on turnsegment for our waiting position
            PathPosition position = turnToElevator.last().clone();
            // but change animation to NONE
            position.animation = Paths.ANIMATE_NONE;
            segments.add(Paths.conditionalSegment(position, waitForComesDown));
            
            // when elevator comes down, walk to elevator
            PathIterator walkToElevator = Paths.linearSegment(firstDirection, from,
                    elevatorSurface, context.velocity * land.cellWidth, Paths.ANIMATE_FORWARD);
            segments.add(walkToElevator);
            
            // this time we wait till elevator moves
            Paths.Condition waitForMoves = new Paths.Condition() {
                public boolean isConditionMet() {
                    return (elevator.timeLeft <= 0);
                }
            };
            // again we use last of previous segment
            position = walkToElevator.last().clone();
            position.animation = Paths.ANIMATE_NONE;
            segments.add(Paths.conditionalSegment(position, waitForMoves));
            
            // another copy of last of walk to elevator segment
            final PathPosition positionAtElevator = walkToElevator.last().clone();
            SimpleVector atElevatorUpstairs = new SimpleVector(elevatorSurface);
            atElevatorUpstairs.y -= elevator.maxHeight;
            
            // we will turn to destination point (next point in path) while going upstairs
            SimpleVector secondDirection = to.calcSub(atElevatorUpstairs);
            final PathIterator turntoLast = Paths.turnSegment(positionAtElevator.direction, secondDirection, atElevatorUpstairs, context.turnVelocity);
            
            // we will move the avatar up with elevator but only till elevator goes up
            // the condition specifies that and also moves the avatar with elavator
            Paths.Condition goUp = new Paths.Condition() {
                public boolean isConditionMet() {
                    positionAtElevator.animation = Land.ACTION_NONE;
                    positionAtElevator.location.y = elevatorSurface.y - elevator.currentHeight;
                    return (elevator.timeLeft > 0);
                }
            };
            // Paths.Wrapper's are typically used for modifiying PathPositions on the fly
            // here we wrap the go up path to turn the avatar while going up
            PathIterator goUpPath = new Paths.Wrapper(Paths.conditionalSegment(positionAtElevator, goUp)) {
                public PathPosition next(float time) {
                    PathPosition position = super.next(time);
                    PathPosition turnPos = turntoLast.next(time);
                    position.direction.set(turnPos.direction);
                    position.animation = turnPos.animation;
                    return position;
                }
            };
            segments.add(goUpPath);
            
            // we are upstairs now, finally walk to destination point
            segments.add(Paths.linearSegment(secondDirection,
                    atElevatorUpstairs, to, context.velocity * land.cellWidth, Paths.ANIMATE_FORWARD));
            
            //that's all, we created all segments. now merge all into a single 
            // path and return it
            return Paths.mergePaths(segments);
        }
    }
    
    
    public static void main(String[] args) throws Exception {
        Logger.setOnError(Logger.ON_ERROR_THROW_EXCEPTION);
        Logger.setLogLevel(Logger.LL_ERRORS_AND_WARNINGS);
        
        // set a few jPCT config variables
        Config.farPlane = 10000;
        Config.linearDiv = 200;
        Config.lightMul = 5;
        
        // create the sample
        ElevatorSample sample = new ElevatorSample();
        
        // pack it in JFrame
        JFrame frame = new JFrame("kilavuz - elevator sample");
        frame.add(sample.renderPanel);
        frame.pack();
        frame.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
        frame.setVisible(true);
        
        sample.renderPanel.requestFocus();
        
        // here we go
        sample.gameLoop();
    }
    
}
